<?php

class MfApi {

	private $url;
	private $username;
	private $password;

	function __construct() {
		$mf_settings = get_option( MOREFLO_API_URL );
		$this->url   = ( isset( $mf_settings['is_active'] ) ) ? $mf_settings['is_active'] : "test";

		//ToDo: Handle the trailing slash
		if ( $this->url == "production" ) {
			$this->url      = trim( $mf_settings['api_production_url'] );
			$this->username = trim( $mf_settings['api_production_user'] );
			$this->password = trim( $mf_settings['api_production_pw'] );
		} else {
			$this->url      = trim( isset( $mf_settings['api_test_url'] ) ?  $mf_settings['api_test_url'] : '' );
			$this->username = trim( isset( $mf_settings['api_test_user'] ) ? $mf_settings['api_test_user'] : '' );
			$this->password = trim( isset( $mf_settings['api_test_pw'] ) ? $mf_settings['api_test_pw'] : '' );
		}
	}


	/*ToDo: Rename this to request once old method depreciated */

	function post_articles( $article ) {

		$body = json_encode( array( $article ) );

		$response = $this->request_new( 'POST', ARTICLES, $body );

		if ( $response ) {
			//Return the new article ID
			return $response;
		}

		return false;
	}

	function request_new( $method, $endpoint, $body = '', $headers = [], $version = 'v2' ) {

		if ( empty( $this->url ) || empty( $this->username ) || empty( $this->password ) ) {
			MfLog::log( ERROR, 'No valid API credentials' );
			return false;
		}

		if ( $version ) {
			$version .= '/';
		}
		$requesturl = sprintf( '%s%s%s/', $this->url, $version, $endpoint );

		$headers['Authorization'] = 'Basic ' . base64_encode( "$this->username:$this->password" );
		$headers['Content-Type']  = 'application/json; charset=utf-8';

		$params = array(
			'method'  => $method,
			'headers' => $headers,
			'timeout' => 30 //ToDo: Move to the config
		);

		if ( ! empty( $body ) && is_string( $body ) ) {
			$params['body'] = $body;
		} else if ( ! empty( $body ) && is_array( $body ) ) {
			$requesturl .= '?' . http_build_query( $body );
		}

		MfLog::log( DEBUG, 'Request start' );
		MfLog::log( DEBUG, $body );

		usleep(200 * 1000); // Quite down api calls, 200 ms sleep
		$response = $this->get_request_response( $requesturl, $params );

		MfLog::log( DEBUG, 'Request end' );

		return $response;
	}


	private function get_request_response( $requesturl, $params, $retry_count = 0 ) {

		$request     = wp_remote_request( esc_url_raw( $requesturl ), $params );
		$status_code = wp_remote_retrieve_response_code( $request );

		if ( is_wp_error( $request ) ) {
			MfLog::log( ERROR, 'START: API Request failed.' );
			MfLog::log( ERROR, $requesturl );
			MfLog::log( ERROR, print_r( $params, true ) );

			$messages = $request->get_error_messages();

			if ( is_array( $messages ) && ! empty( $messages ) ) {
				MfLog::log( ERROR, print_r( $messages, true ) );
			}

			MfLog::log( ERROR, 'END: API Request failed.' );
		} else {
			if ( isset( $request['body'] ) && $status_code == 200 ) {
				MfLog::log( DEBUG, $request['body'] );

				return $request['body'];
			} else {
				MfLog::log( ERROR, 'START: API Request failed.' );
				MfLog::log( ERROR, $requesturl );
				MfLog::log( ERROR, print_r( $params, true ) );
				MfLog::log( ERROR, 'END: API Request failed.' );
			}
		}

		$retry_limit = 5; //ToDo: Move to the config?

		$retry_codes = array( // Array of request response codes that should trigger retry.
			429 => 'API Rate limit reached',
		);

		if ( ! empty( $retry_codes[ $status_code ] ) ) {
			if ( $retry_count <= $retry_limit ) {
				MfLog::log( ERROR, $retry_codes[ $status_code ] . ', retrying after short pause...' );
				usleep(1000 * 1000);
				$retry_count++;
				return $this->get_request_response( $requesturl, $params, $retry_count );
			} else {
				MfLog::log( ERROR, 'Request retry limit reached. Aborting request.' );
				return false;
			}
		}

		return false;
	}

	/*
	 * Rename this article
	 */

	function get_articles_new( $params ) {

		if ( empty( $params['page'] ) ) {
			$params['page'] = 0;
		}

		if ( empty( $params['pageSize'] ) ) {
			$params['pageSize'] = 100;
		}

		$response = $this->request_new( 'GET', ARTICLES, $params );

		if ( $response ) {
			return json_decode( $response, true );
		}

		return false;
	}

	function put_articles( $article ) {

		$mf_settings     = get_option( MOREFLO_API_URL );
		if ( $mf_settings[ISPRICEACTIVE] != 'enabled' ) {
			MfLog::log( DEBUG, 'Price sync disabled.' );
			unset( $article[RECOMMENDEDPRICE] );
			unset( $article[RETAILPRICE] );
		}

		if ( isset( $article[QUANTITY] ) ) {
			unset( $article[QUANTITY] );
		}


		$body = json_encode( array( $article ) );

		$response = $this->request_new( 'PUT', ARTICLES, $body );

		if ( $response ) {
			//Return the new article ID
			return $response;
		}

		return false;
	}

	function get_article( $articleNo ) {
		$response = $this->request_new( 'GET', 'articles/' . $articleNo );
		if ( $response ) {
			return $response;
		}
		return false;
	}

	function get_article_sets( $params ) {
		if ( empty( $params['page'] ) ) {
			$params['page'] = 0;
		}

		if ( empty( $params['pageSize'] ) ) {
			$params['pageSize'] = 100;
		}

		$response = $this->request_new( 'GET', 'articlesets', $params );

		if ( $response ) {
			return json_decode( $response, true );
		}

		return false;
	}

	function get_account_codes() {

		$response = $this->request_new( 'GET', 'accountcodes' );

		if ( $response ) {
			return $response;
		}

		return false;
	}

	function get_order( $id ) {

		$response = $this->request_new( 'GET', 'orders/byid/' . $id, '', [], 'v3');

		if ( $response ) {
			return $response;
		}

		return false;
	}

	function get_service_order( $id ) {

		$response = $this->request_new( 'GET', 'serviceorders/' . $id);

		if ( $response ) {
			return $response;
		}

		return false;
	}

	function get_orders( $modified_since, $page, $page_size ) {
		$params = [
			'modifiedSince' => $modified_since,
			'page' => $page,
			'pageSize' => $page_size,
		];

		$response = $this->request_new( 'GET', 'orders', $params, [], 'v3' );

		if ( $response ) {
			return json_decode( $response, true );
		}

		return false;
	}

	function get_customer($customer)
	{
		$body = json_encode($customer);

		$response = $this->request_new( 'POST', 'customers/find', $body );

		if ( $response ) {
			//Return customer ID
			return json_decode($response, true);
		}

		return false;
	}

	function post_order( $order ) {

		$body = json_encode( array( $order ) );

		$response = $this->request_new( 'POST', 'orders', $body, [], 'v3' );

		if ( $response ) {
			//Return the new order ID
			return $response;
		}

		return false;
	}

	function post_service_order( $order ) {

		$body = json_encode( array( $order ) );

		$response = $this->request_new( 'POST', 'serviceorders', $body );

		if ( $response ) {
			//Return the new order ID
			return $response;
		}

		return false;
	}

	function post_receipt( $receipt ) {

		$body = json_encode( $receipt );
		$response = $this->request_new( 'POST', 'receipts', $body, [] );

		if ( $response ) {
			//Return the new order ID
			return $response;
		}

		return false;
	}


	function put_decrease_stock( $items ) {

		$body = json_encode( $items );

		$response = $this->request_new( 'PUT', 'articles/decreasestock', $body );

		if ( $response ) {
			//Return articles
			return $response;
		}

		return false;
	}

	function put_increase_stock( $items ) {

		$body = json_encode( $items );

		$response = $this->request_new( 'PUT', 'articles/increasestock', $body );

		if ( $response ) {
			//Return articles
			return $response;
		}

		return false;
	}

	function get_image_file($image_url, $image_type)
	{
		if (empty( $this->username ) || empty( $this->password ) ) {
			MfLog::log( ERROR, 'No valid API credentials' );
			return false;
		}
		
		$context = stream_context_create(array(
			"http" => array(
				"header" => "Authorization: Basic " . base64_encode("$this->username:$this->password"),
				"Content-Type"  => $image_type
			)));

		$data = file_get_contents($image_url, false, $context);
		return $data;
	}

	function get_groups( $params ) {

		if ( empty( $params['page'] ) ) {
			$params['page'] = 0;
		}

		if ( empty( $params['pageSize'] ) ) {
			$params['pageSize'] = 1000;
		}

		$response = $this->request_new( 'GET', 'groups', $params );

		if ( $response ) {
			return json_decode( $response, true );
		}

		return false;
	}

	function get_campaigns( $params ) {

		if ( empty( $params['page'] ) ) {
			$params['page'] = 0;
		}

		if ( empty( $params['pageSize'] ) ) {
			$params['pageSize'] = 100;
		}

		$response = $this->request_new( 'GET', 'campaigns', $params );

		if ( $response ) {
			return json_decode( $response, true );
		}

		return false;
	}

	function get_vat_accounts( $params ) {

		if ( empty( $params['page'] ) ) {
			$params['page'] = 0;
		}

		if ( empty( $params['pageSize'] ) ) {
			$params['pageSize'] = 100;
		}

		$response = $this->request_new( 'GET', 'settings/accounting/vat', $params );

		if ( $response ) {
			return json_decode( $response, true );
		}

		return false;
	}
}
