<?php

defined( 'ABSPATH' ) || exit;

class MF_Sync_Campaigns {
    public function __construct() {
        if ( ! moreflo_sync_campaigns_enabled() ) {
            return;
        }

        add_filter( 'woocommerce_coupon_discount_types', [ $this, 'coupon_discount_types' ] );
        add_filter( 'woocommerce_cart_coupon_types', [ $this, 'cart_coupon_types' ] );
        add_filter( 'woocommerce_coupon_get_discount_amount', [ $this, 'get_discount_amount' ], 10, 5 );
		add_action( 'woocommerce_before_cart', [ $this, 'update_cart_discounts' ] );
		add_action( 'woocommerce_before_checkout_form', [ $this, 'update_cart_discounts' ] );
		add_action( 'woocommerce_applied_coupon', [ $this, 'remove_wrongly_applied_discounts' ], 10, 1 );
		add_filter( 'woocommerce_add_message', [ $this, 'filter_woocommerce_add_message' ] );
		add_filter( 'woocommerce_cart_totals_coupon_label', [ $this, 'coupon_label' ], 20, 2 );

		add_action( 'moreflo_sync_update_cart_discounts', [ $this, 'update_cart_discounts' ] );
    }

	/**
	 * Hooks into woocommerce_coupon_discount_types.
	 * 
	 * Adds Apps coupon types to WooCommerce
	 * 
	 * @param array $discount_types
	 * @return array
	 */
    public function coupon_discount_types( $discount_types ) {
        $discount_types['mf_type_0'] = __( 'Buy X amount and get Y amount', 'wpmf' );
        $discount_types['mf_type_1'] = __( 'Buy X amount and get Y discount', 'wpmf' );
        $discount_types['mf_type_2'] = __( 'Buy X amount and pay Y', 'wpmf' );
        $discount_types['mf_type_3'] = __( 'Percent Discount', 'wpmf' );
        $discount_types['mf_type_4'] = __( 'Special Offer', 'wpmf' );
        $discount_types['mf_type_5'] = __( 'Price Reduction', 'wpmf' );

        $discount_types['mf_type_set'] = __( 'Article Set', 'wpmf' );

        return $discount_types;
    }

	/**
	 * Hooks into woocommerce_cart_coupon_types.
	 * 
	 * Adds Apps coupon types to WooCommerce cart
	 * 
	 * @param array $coupon_types
	 * @return array
	 */
    public function cart_coupon_types( $coupon_types ) {
        $coupon_types[] = 'mf_type_0';
        $coupon_types[] = 'mf_type_1';
        $coupon_types[] = 'mf_type_2';
        $coupon_types[] = 'mf_type_3';
        $coupon_types[] = 'mf_type_4';
        $coupon_types[] = 'mf_type_5';

        $coupon_types[] = 'mf_type_set';

        return $coupon_types;
    }

    /**
	 * Hooks into woocommerce_coupon_get_discount_amount.
	 * 
     * Gets the discounted amout for a specific cart item.
	 * 
	 * @param float 	 $discount Current discounted amout.
	 * @param float 	 $discounting_amount Amount the coupon is being applied to.
	 * @param array|null $cart_item Cart item being discounted if applicable.
	 * @param boolean 	 $single True if discounting a single qty item, false if its the line.
	 * @param WC_Coupon  $coupon The current coupon being applied.
	 * @return float New discounted amount.
     */
    public function get_discount_amount( $discount, $discounting_amount, $cart_item, $single, $coupon ) {
		// We cannot provide discount when no cart item is specified
		if ( is_null( $cart_item ) ) {
			return $discount;
		}

        $products = $this->get_products_from_cart( WC()->cart );

		// If there are no products in the cart, abort
        if ( empty( $products ) ) {
            return $discount;
        }

		$discounter = MF_Sync_Campaigns_Factory::create_discounter( $coupon );

		// Not an Apps coupon
		if ( is_null( $discounter ) ) {
			return $discount;
		}

		// Calulate discount
		$discount = $this->calc_discount( $discounter, $cart_item, $products, WC()->cart->subtotal );

		return mf_wc_round( $discount, wc_get_rounding_precision() );
    }

	/**
	 * Updates coupons in cart based on it's contents.
	 */
	public function update_cart_discounts() {
		$discounters = $this->get_cart_discounters();

		$this->toggle_wc_notices(false);
		
		// Adds coupons to cart if it's not already added
		if ( ! empty( $discounters ) ) {
			foreach ( $discounters as $discounter ) {
				$coupon_code = $discounter->get_coupon()->get_code();

				// Do not apply the same coupon twice
				if ( WC()->cart->has_discount( $coupon_code ) ) {
					// Cart already contains discounter
					continue;
				}

				WC()->cart->apply_coupon( $coupon_code );
			}
		}

		$this->toggle_wc_notices(true);

		// Remove discounts not applicable anymore (keep applicable discounts)
		$this->clear_cart_of_discounts( $discounters );
	}

	public function remove_wrongly_applied_discounts( $coupon_code ) {
		$discounters = $this->get_cart_discounters();
	
		// Remove discounts not applicable anymore (keep applicable discounts)
		$this->clear_cart_of_discounts( $discounters );
	}

	public function filter_woocommerce_add_message( $message ) {
		return $this->wc_notices
			? $message
			: '';
	}

	/**
	 * @param string $label
	 * @param WC_Coupon $coupon
	 */
	public function coupon_label( $label, $coupon ) {
		$mf_coupon_types = $this->cart_coupon_types([]);

		if ( in_array( $coupon->get_discount_type(), $mf_coupon_types ) ) {
			$label = sprintf( esc_html__( 'Campaign: %s', 'wpmf' ), $coupon->get_code() );
		}

		return $label;
	}

	/**
	 * Toggles adding of WooCommerce messages on or off
	 * 
	 * @param bool $on
	 */
	private function toggle_wc_notices( $on ) {
		$this->wc_notices = $on;
	}

	private $wc_notices = true;

	/**
	 * @return MF_Campaigns_Discounter[]
	 */
	private function get_cart_discounters() {
		// We need a cart to push coupons onto
		if ( is_null( WC()->cart ) ) {
			return [];
		}

		// Get all products from cart
        $products = $this->get_products_from_cart( WC()->cart );

		// Get possible discounts
		$possible_discounts = array_filter(
			$this->get_available_discounts_for_cart( $products ),
			[ $this, 'filter_active_now' ]
		);

		// Keep a list of applicable discounts so we can remove everything else
		$applicable_discounts = [];

		// Adds coupons to cart if it's not already added
		if ( ! empty( $possible_discounts ) ) {
			foreach ( $possible_discounts as $possible_discount ) {
				// Check if coupon is applicable for the cart
				if ( ! $possible_discount->is_applicable( WC()->cart->subtotal, $products ) ) {
					continue;
				}

				// Get the WooCommerce coupon
				$coupon_code = $possible_discount->get_coupon()->get_code();
				if ( ! $coupon_code ) {
					return;
				}

				// Keep a list of applicable discounts
				$applicable_discounts[] = $possible_discount;
			}
		}

		return $applicable_discounts;
	}

	/**
	 * Removes all discounts except the applicable ones from the cart.
	 * 
	 * @param MF_Campaigns_Discounter[] $applicable_discounts
	 */
	private function clear_cart_of_discounts( $applicable_discounts ) {
		// We need a cart
		if ( is_null( WC()->cart ) ) {
			return;
		}

		$cart_coupons = WC()->cart->get_coupons();

		// Cart already cleared of coupons
		if ( empty( $cart_coupons ) ) {
			return;
		}

		$site_coupons = $this->get_possible_site_apps_coupons();
		// No coupons to remove
		if ( empty( $site_coupons ) ) {
			return;
		}

		$site_coupon_codes = array_map(
			function ( $coupon ) {
				return $coupon->get_code();
			},
			$site_coupons
		);

		$applicable_discounts_codes = array_map(
			function ( $discount ) {
				return $discount->get_coupon()->get_code();
			},
			$applicable_discounts
		);

		$this->toggle_wc_notices(false);

		// Loop through site coupons and remove them from the cart if they have been applied
		foreach ( $cart_coupons as $cart_coupon ) {
			// Do not remove applicable discounts
			if ( in_array( $cart_coupon->get_code(), $applicable_discounts_codes ) ) {
				continue;
			}

			// Do not remove discounts not found in cart
			if ( ! in_array( $cart_coupon->get_code(), $site_coupon_codes ) ) {
				continue;
			}

			WC()->cart->remove_coupon( $cart_coupon->get_code() );
		}

		$this->toggle_wc_notices(true);
	}

	/**
	 * Gets available discounts for given products.
	 * 
	 * @param MF_Campaigns_Product[] $products
	 * @return MF_Campaigns_Discounter[]
	 */
	private function get_available_discounts_for_cart( $products ) {
		if ( empty( $products ) ) {
			return [];
		}

		$coupons = $this->get_possible_site_apps_coupons();

		$skus = array_unique( array_map( function ( $product ) {
			return $product->get_sku();
		}, $products ) );

		$coupons = array_filter( $coupons, function ( $coupon ) use ( $skus ) {
			$coupon_skus = json_decode( get_post_meta( $coupon->get_id(), 'mf_parts_sku', true ) );
			if ( is_null( $coupon_skus ) ) {
				// Article sets lacks parts sku
				return true;
			}

			foreach ( $skus as $sku ) {
				if ( in_array( $sku, $coupon_skus ) ) {
					return true;
				}
			}

			return false;
		} );

		if ( empty( $coupons ) ) {
			return [];
		}

		$discounters = array_filter(
			array_map(
				function ( $coupon ) {
					return MF_Sync_Campaigns_Factory::create_discounter( $coupon );
				},
				$coupons
			)
		);

		usort( $discounters, [ $this, 'sort_by_priority' ] );

		return $discounters;
	}

	/**
	 * Gets all site coupons that originated from Apps.
	 * 
	 * @return WC_Coupon[]
	 */
	private function get_possible_site_apps_coupons() {
		$posts = get_posts([
			'post_type' => 'shop_coupon',
			'post_status' => 'publish',
			'ignore_sticky_posts' => true,
			'nopaging' => true,
			'meta_query' => [
				[
					'key' => 'mf_id',
					'compare' => 'EXISTS',
				],
				[
					'key' => 'mf_end_date',
					'compare' => '>=',
					'value' => (new DateTime())->format( 'Y-m-d' ),
				]
			]
		]);

		return array_map(
			function ( $post ) {
				return new WC_Coupon( $post->post_title );
			},
			$posts
		);
	}

	/**
	 * Calculate cart item per quantity discount with discounter.
	 * 
	 * @param MF_Campaigns_Discounter $discounter
	 * @param array $cart_item
	 * @param MF_Campaigns_Product[] $products
	 * @param float $cart_total Total price of cart, inc VAT
	 * @return float
	 */
    private function calc_discount( $discounter, $cart_item, $products, $cart_total ) {
        $discounted_products = $discounter->get_discounted_products( $products, $cart_total );
		
		if ( empty( $discounted_products ) ) {
			return 0;
		}

		$total_discount = 0;
		foreach ( $discounted_products as $discounted_product ) {
			if ( $cart_item['data']->get_sku() != $discounted_product->get_sku() ) {
				continue;
			}

			$total_discount += $discounted_product->get_discount( ! wc_prices_include_tax() );
		}

		$per_quantity_discount = $total_discount / $cart_item['quantity'];

        return $per_quantity_discount;
    }

	/**
	 * Get products from the cart.
	 *
	 * @param WC_Cart $cart Cart object.
	 * @return MF_Campaigns_Product[] Products ordered by price ASC. All prices are inc VAT.
	 */
	private function get_products_from_cart( $cart ) {
		if ( ! is_a( $cart, 'WC_Cart' ) ) {
			return array();
		}

		$products = array();

		// Transform cart items to products
		foreach ( $cart->get_cart_contents() as $cart_item ) {
			$cart_item_products = MF_Sync_Campaigns_Factory::create_products_from_cart_item( $cart_item, $cart );

			// We get a product per quantity of item
			foreach ( $cart_item_products as $cart_item_product ) {
				$products[] = $cart_item_product;
			}
		}

        usort( $products, array( $this, 'sort_by_price' ) );

        return $products;
	}

	/**
	 * Sort by price ascending.
	 *
	 * @param MF_Campaigns_Product $a
	 * @param MF_Campaigns_Product $b
	 * @return int
	 */
	private function sort_by_price( $a, $b ) {
		if ( $a->get_price() === $b->get_price() ) {
			return 0;
		}

		return ( $a->get_price() < $b->get_price() ) ? -1 : 1;
	}

	/**
	 * Sort by priority ascending.
	 * 
	 * @param MF_Campaigns_Discounter $a
	 * @param MF_Campaigns_Discounter $b
	 * @return int
	 */
	private function sort_by_priority( $a, $b ) {
		if ( $a->get_priority() === $b->get_priority() ) {
			return 0;
		}

		return ( $a->get_priority() < $b->get_priority() ) ? -1 : 1;
	}

	/**
	 * Filters discounters so that those within start and end times are left.
	 * 
	 * @param MF_Campaigns_Discounter $discounter
	 * @return bool
	 */
	private function filter_active_now( $discounter ) {
		$now = new DateTime();
		$start_time = DateTime::createFromFormat( 'Y-m-d\\TH:i:s', get_post_meta( $discounter->get_coupon()->get_id(), 'mf_start_time', true ), new DateTimeZone( 'Europe/Stockholm' ) );
		$end_time = DateTime::createFromFormat( 'Y-m-d\\TH:i:s', get_post_meta( $discounter->get_coupon()->get_id(), 'mf_end_time', true ), new DateTimeZone( 'Europe/Stockholm' ) );

		// Remove passed discounts
		if ( $end_time && $now > $end_time ) {
			return false;
		}

		// Remove future discounts
		if ( $start_time && $now < $start_time ) {
			return false;
		}
		
		return true;
	}
}

new MF_Sync_Campaigns();
