<?php

defined( 'ABSPATH' ) || exit;

class MF_Campaigns_Type_Set implements MF_Campaigns_Discounter {
    /**
     * @param WC_Coupon $coupon
     * @param array $values
     * @param int $priority
     */
    public function __construct( $coupon, $values, $priority ) {
        $this->coupon = $coupon;
        $this->values = $values;
        $this->priority = intval( $priority );
    }

    public function get_name() {
        return __( 'Article Set', 'wpmf' );
    }

    public function get_coupon() {
        return $this->coupon;
    }

    public function get_priority() {
        return $this->priority;
    }

    public function is_applicable( $cart_total, $products ) {
        if ( empty( $this->values ) ) {
            return false;
        }

        $applicable_products = $this->applicable_products( $products );

        $counter = [];
        foreach ( $this->values as $v ) {
            $counter[ $v->sku ] = [
                'needed' => $v->quantity, // Needed quantity for be valid
                'current' => 0, // Counter
            ];
        }

        foreach ( $applicable_products as $applicable_product ) {
            $counter[ $applicable_product->get_sku() ]['current']++;
        }

        foreach ( $counter as $c ) {
            if ( $c['current'] < $c['needed'] ) {
                return false;
            }
        }

        return true;
    }

    public function get_discounted_products( $products, $cart_total ) {
        return $this->_get_discounted_products(
            $this->applicable_products( $products ),
            $cart_total
        );
    }

    protected function applicable_products( $products ) {
        $values = $this->values;
        return array_filter(
            $products,
            function ( $product ) use ( $values ) {
                if ( $product->get_uses() > 0 ) {
                    return false;
                }
                foreach ( $values as $v ) {
                    if ( $v->sku != $product->get_sku() ) {
                        continue;
                    }

                    return true;
                }

                return false;
            }
        );
    }

    /**
     * @param MF_Campaigns_Product[] $products
     * @param int $maximum_allowed_discount
     * @return MF_Campaigns_Product[]
     */
    protected function _get_discounted_products( $products, $maximum_allowed_discount ) {
        if ( count( $products ) < 1 ) {
            return [];
        }

        if ( ! $this->is_applicable( $maximum_allowed_discount, $products ) ) {
            return $products;
        }

        $products = array_values( $products );

        $discounted_products = [];

        foreach ( $this->values as $v ) {
            for ( $i = 0; $i < $v->quantity; $i++ ) {
                $key = null;

                foreach ( $products as $k => $p ) {
                    if ( $p->get_sku() == $v->sku ) {
                        $key = $k;
                        break;
                    }
                }

                if ( is_null( $key ) ) {
                    // We don't have correct quantity of some product, abort
                    return array_merge(
                        $discounted_products,
                        $products
                    );
                }

                $discounted_product = array_splice( $products, $key, 1 )[0];
                $discounted_product->set_discount( $discounted_product->get_price() - $v->price_per_unit );
                $discounted_products[] = $discounted_product;
            }
        }

        if ( empty( $products ) ) {
            return array_merge(
                $discounted_products
            );
        }

        return array_merge(
            $discounted_products,
            $this->_get_discounted_products( $products, $maximum_allowed_discount )
        );
    }

    /** @var WC_Coupon WooCommerce coupon for this discount */
    protected $coupon;
    /** @var array List of articles with quantity, sku and price per unit */
    protected $values;
    /** @var int Priority */
    protected $priority;
}