<?php

defined( 'ABSPATH' ) || exit;

class MF_Sync_Campaigns_Sync_Sets implements MF_Sync_Sync {
    /**
     * Syncs article sets to WooCommerce as coupons
     */
    public function sync( $page, $pageSize, $updatedSince ) {
		$api = new MfApi();
		$status = array('delete' => 0, 'ignore' => 0, 'error' => 0, 'add' => 0, 'new' => 0, 'update' => 0, 'page' => 0, 'total' => 1);
		$params = array(
			'page' => $page,
			'pageSize' =>  $pageSize
		);

        if ( $updatedSince ) {
            $params['UpdatedSince'] = $updatedSince;
        }

		$sets = $api->get_article_sets($params);

		if(!empty($sets) && !empty($sets['Total'])) {
			$status['page'] = $sets['Page'];
			$status['total'] = $sets['Total'];
		}

		if(!empty($sets['list']) && is_array($sets['list'])) {
            $status = $this->process_pulled_sets( $sets['list'], $status );
		}

		return $status;
    }

	public function process_pulled_sets( $list, $status ) {
		if ( empty( $list ) ) {
			return $status;
		}

        foreach ( $list as $set ) {
            if(empty($set['ArticleSet'])) {
                continue;
            }
            
            $action = $this->process_pulled_set($set['ArticleSet']);

            if($action !== false) {
                $status[$action]++;
            }
        }

		return $status;
	}

	public function process_pulled_set($set) {
		$post_type = 'shop_coupon';

        if ( ! $set['IsActive'] ) {
			MfLog::log(DEBUG, 'Article set is deleted or disabled. ('. $set['ArticleNumber'] .')');
            $deleted = $this->delete_post( $set['ArticleNumber'], $post_type );
            return $deleted
                ? 'delete'
                : 'ignore';
        }

		MfLog::log(DEBUG, 'Processing set: ' . $set['ArticleNumber']);

        $post = $this->update_post( $set['ArticleNumber'], $set['ArticleName'], $post_type );

        if ( ! $post ) {
            return 'error';
        }

		// Add meta
		update_post_meta( $post, 'discount_type', 'mf_type_set' );
		update_post_meta( $post, 'mf_priority', 0 ); // Sets are always highest priority

        $parts = $this->get_parts( $set['articleSetParts'] );
        update_post_meta( $post, 'mf_parts', json_encode( $parts ) );

		update_post_meta( $post, 'mf_end_date', (new DateTime())->add( new DateInterval( 'P1Y' ) )->format( 'Y-m-d' ) ); // Should always be active
		update_post_meta( $post, 'mf_last_updated', time() );

		MfLog::log(DEBUG, 'Processed campaign: ' . $set['ArticleNumber']);

		return 'add';
	}

    protected function get_parts( $parts ) {
        $r = [];

        if ( empty( $parts ) ) {
            return $r;
        }

        foreach ( $parts as $part ) {
            if ( 0 == $part['PartType'] ) {
                // Single SKU
                $r[] = [
                    'sku' => $part['ArticleReference'],
                    'quantity' => $part['Quantity'],
                    'price_per_unit' => $part['PartPricePerUnit'],
                ];
            }
        }

        return $r;
    }

    /**
     * Creates or updates post that has unique mf id
     * 
     * @param string $mf_id Unique mf id for this post
     * @param string $name Title of the post
     * @param string $post_type Post type
     * @param string $content Post Content
     * @return int|bool Post ID on success, else false
     */
    protected function update_post( $mf_id, $name, $post_type, $content = '' ) {
        $post_id = $this->get_post_id( $mf_id, $post_type );

        if ( $post_id ) {
            $post = get_post( $post_id );
            $updates = [];
            if ( $name != $post->post_title ) {
                $updates['post_title'] = $name;
            }
            if ( $content != $post->post_content ) {
                $updates['post_content'] = $content;
            }
            if ( ! empty( $updates ) ) {
                $updates['ID'] = $post_id;
                wp_update_post( $updates );
				MfLog::log('debug', 'Updated post in WP ('. $mf_id .'; '. $post_type .'; '. json_encode( $updates ) .')' );
            }
        } else {
            $post_id = wp_insert_post([
                'post_type' => $post_type,
				'post_author' => 1,
                'post_title' => $name,
                'post_status' => 'publish',
                'post_content' => $content,
            ]);

            if ( is_wp_error( $post_id ) ) {
				MfLog::log('error', 'Error creating new post in WP ('. $mf_id .'): '. $post_id->get_error_message() );
                return false;
            }

            if ( ! $post_id ) {
				MfLog::log('error', 'Error creating new post in WP ('. $mf_id .'): wp_insert_post returned 0' );
                return false;
            }

            update_post_meta( $post_id, 'coupon_amount', '' );
            update_post_meta( $post_id, 'individual_use', 'no' );
            update_post_meta( $post_id, 'product_ids', '' );
            update_post_meta( $post_id, 'exclude_product_ids', '' );
            update_post_meta( $post_id, 'usage_limit', '' );
            update_post_meta( $post_id, 'expiry_date', '' );
            update_post_meta( $post_id, 'apply_before_tax', 'yes' );
            update_post_meta( $post_id, 'free_shipping', 'no' );

            update_post_meta( $post_id, 'mf_id', $mf_id );
            update_post_meta( $post_id, 'mf_type', 'article_set' );
			MfLog::log('debug', 'New post created in WP ('. $mf_id .'; '. $post_type .'; '. $name .')' );
        }
        
        return $post_id;
    }

    /**
     * Gets post id that has unique mf id.
     * 
     * @param string $mf_id Unique mf id
     * @param string $post_type Post Type
     * @return int|bool Post id on success, else false
     */
    protected function get_post_id( $mf_id, $post_type ) {
        $posts = get_posts([
            'post_type' => $post_type,
            'status' => 'all',
            'posts_per_page' => 1,
            'ignore_sticky_posts' => true,
            'meta_query' => [
                [
                    'key' => 'mf_id',
                    'value' => $mf_id,
                ],
                [
                    'key' => 'mf_type',
                    'value' => 'article_set',
                ]
            ],
        ]);

        if ( ! empty( $posts ) ) {
            return $posts[0]->ID;
        }

        return false;
    }

    /**
     * Deletes a post that has unique mf id. If it's not found then it does nothing
     * 
     * @param string $mf_id Unique mf id
     * @param string $post_type Post type
     * @return bool True on deleted, false on not found
     */
    protected function delete_post( $mf_id, $post_type ) {
        $post_id = $this->get_post_id( $mf_id, $post_type );

        if ( $post_id ) {
            $title = get_the_title( $post_id );
            wp_delete_post( $post_id );
			MfLog::log('debug', 'Post deleted in WP ('. $mf_id .'; '. $post_type .'; '. $title .')' );
            return true;
        }

        return false;
    }
}
