<?php

defined( 'ABSPATH' ) || exit;

class MF_Sync_Campaigns_Sync implements MF_Sync_Sync {
    /**
     * Syncs campaigns to WooCommerce as coupons
     */
    public function sync( $page, $pageSize, $updatedSince ) {
		$api = new MfApi();
		$status = array('delete' => 0, 'ignore' => 0, 'error' => 0, 'add' => 0, 'new' => 0, 'update' => 0, 'page' => 0, 'total' => 1);
		$params = array(
			'page' => $page,
			'pageSize' =>  $pageSize
		);

		$campaigns = $api->get_campaigns($params);

		if(!empty($campaigns) && !empty($campaigns['Total'])) {
			$status['page'] = $campaigns['Page'];
			$status['total'] = $campaigns['Total'];
		}

		if(!empty($campaigns['list']) && is_array($campaigns['list'])) {
			$status = $this->process_pulled_campaigns( $campaigns['list'], $status );
		}

		return $status;
    }

	public function process_pulled_campaigns( $list, $status ) {
		if ( empty( $list ) ) {
			return $status;
		}

        foreach ( $list as $campaign ) {
            if(empty($campaign['Campaign'])) {
                continue;
            }
            
            $action = $this->process_pulled_campaign($campaign['Campaign']);

            if($action !== false) {
                $status[$action]++;
            }
        }

		return $status;
	}

	public function process_pulled_campaign($campaign) {
		$post_type = 'shop_coupon';

        if ( ! $campaign['Enabled'] ) {
			MfLog::log(DEBUG, 'Campaign is deleted or disabled. ('. $campaign['Id'] .')');
            $deleted = $this->delete_post( $campaign['Id'], $post_type );
            return $deleted
                ? 'delete'
                : 'ignore';
        }

        if ( $campaign['MemberOnly'] ) {
			MfLog::log(DEBUG, 'Campaign with member only is not supported. ('. $campaign['Id'] .')');
            $deleted = $this->delete_post( $campaign['Id'], $post_type );
            return $deleted
                ? 'delete'
                : 'ignore';
        }

        if ( $campaign['MemberGroupName'] ) {
			MfLog::log(DEBUG, 'Campaign for member group only is not supported. ('. $campaign['Id'] .')');
            $deleted = $this->delete_post( $campaign['Id'], $post_type );
            return $deleted
                ? 'delete'
                : 'ignore';
        }

        $end_time = DateTime::createFromFormat( 'Y-m-d\\TH:i:s', $campaign['EndTime'] );

        if ( is_null( $end_time ) || ! $end_time || new DateTime() > $end_time ) {
			MfLog::log(DEBUG, 'Campaign has ended. ('. $campaign['Id'] .')');
            $deleted = $this->delete_post( $campaign['Id'], $post_type );
            return $deleted
                ? 'delete'
                : 'ignore';
        }

		MfLog::log(DEBUG, 'Processing campaign: ' . $campaign['Id']);

        $post = $this->update_post( $campaign['Id'], $campaign['Name'], $post_type );

        if ( ! $post ) {
            return 'error';
        }

		// Add meta
		update_post_meta( $post, 'discount_type', 'mf_type_'. $campaign['Type'] );
		update_post_meta( $post, 'mf_value_x', $campaign['ValueX'] );
		update_post_meta( $post, 'mf_value_y', $campaign['ValueY'] );
		update_post_meta( $post, 'mf_value', $campaign['Value'] );
		update_post_meta( $post, 'mf_value_is_global', $campaign['ValueIsGlobal'] );
		update_post_meta( $post, 'mf_priority', $campaign['Priority'] );
		update_post_meta( $post, 'mf_threshold_total', $campaign['ThresholdTotal'] );
		update_post_meta( $post, 'mf_threshold_total_include_this', $campaign['ThresholdTotalIncludeThis'] );
		update_post_meta( $post, 'mf_start_time', $campaign['StartTime'] );
		update_post_meta( $post, 'mf_end_time', $campaign['EndTime'] );

        $parts = $this->get_parts( $campaign['parts'], $campaign['ValueIsGlobal'], $campaign['Value'] );
        update_post_meta( $post, 'mf_parts_sku', json_encode( $parts['sku'] ) );
        update_post_meta( $post, 'mf_parts_value', json_encode( $parts['value'] ) );

		update_post_meta( $post, 'mf_end_date', $end_time->add( new DateInterval( 'P1D' ) )->format( 'Y-m-d' ) );
		update_post_meta( $post, 'mf_last_updated', time() );

		MfLog::log(DEBUG, 'Processed campaign: ' . $campaign['Id']);

		return 'add';
	}

	public function clean_campaigns($page) {
		$status = array('page' => $page, 'total' => 1);

		if ($page > 0) {
			// Already cleaned
			MfLog::log(DEBUG, 'Skipped double clean campaigns');
			return $status;
		}

		$post_type = 'shop_coupon';

		MfLog::log(DEBUG, 'Processing clean campaigns');

		$posts = get_posts([
			'post_type' => $post_type,
			'ignore_sticky_posts' => true,
			'fields' => 'ids',
			'meta_query' => [
				'relation' => 'AND',
				[
					'key' => 'mf_id',
					'compare' => 'EXISTS',
				],
                [
                    'key' => 'mf_type',
                    'value' => 'campaign',
                ],
				[
					'key' => 'mf_last_updated',
					'value' => time() - MINUTE_IN_SECONDS,
					'compare' => '<',
				],
			],
		]);

		if (!empty($posts)) {
			foreach ($posts as $post_id) {
				$mf_id = get_post_meta( $post_id, 'mf_id', true );
				$this->delete_post( $mf_id, $post_type );
			}
		}

		MfLog::log(DEBUG, 'Processed clean campaigns');

		return $status;
	}

    protected function get_parts( $parts, $value_is_global, $global_value ) {
        $r = [
            'sku' => [],
            'value' => [],
        ];

        if ( empty( $parts ) ) {
            return $r;
        }

        foreach ( $parts as $part ) {
            $part_value = $value_is_global
                ? $global_value
                : $part['Value'];
            if ( 0 == $part['PartType'] ) {
                // Single SKU
                if ( ! in_array( $part['Reference'], $r['sku'] ) ) {
                    $r['sku'][] = $part['Reference'];
                }
                $r['value'][ $part['Reference'] ] = $part_value;
            } elseif ( 2 == $part['PartType'] ) {
                // Category
                $query = new WC_Product_Query([
                    'post_status' => 'publish',
                    'ignore_sticky_psots' => true,
                    'nopaging' => true,
                    'tax_query' => [
                        [
                            'taxonomy' => 'product_cat',
                            'field' => 'name',
                            'terms' => $part['Reference'],
                            'include_children' => true,
                        ],
                    ],
                ]);
                $products = $query->get_products();

                if ( ! empty( $products ) ) {
                    $skus = array_map( function ( $product ) {
                        return $product->get_sku();
                    }, $products );
                    foreach ( $skus as $sku ) {
                        if ( ! in_array( $sku, $r['sku'] ) ) {
                            $r['sku'][] = $sku;
                        }
                        $r['value'][ $sku ] = $part_value;
                    }
                }
            }
        }

        return $r;
    }

    /**
     * Creates or updates post that has unique mf id
     * 
     * @param string $mf_id Unique mf id for this post
     * @param string $name Title of the post
     * @param string $post_type Post type
     * @param string $content Post Content
     * @return int|bool Post ID on success, else false
     */
    protected function update_post( $mf_id, $name, $post_type, $content = '' ) {
        $post_id = $this->get_post_id( $mf_id, $post_type );

        if ( $post_id ) {
            $post = get_post( $post_id );
            $updates = [];
            if ( $name != $post->post_title ) {
                $updates['post_title'] = $name;
            }
            if ( $content != $post->post_content ) {
                $updates['post_content'] = $content;
            }
            if ( ! empty( $updates ) ) {
                $updates['ID'] = $post_id;
                wp_update_post( $updates );
				MfLog::log('debug', 'Updated post in WP ('. $mf_id .'; '. $post_type .'; '. json_encode( $updates ) .')' );
            }
        } else {
            $post_id = wp_insert_post([
                'post_type' => $post_type,
				'post_author' => 1,
                'post_title' => $name,
                'post_status' => 'publish',
                'post_content' => $content,
            ]);

            if ( is_wp_error( $post_id ) ) {
				MfLog::log('error', 'Error creating new post in WP ('. $mf_id .'): '. $post_id->get_error_message() );
                return false;
            }

            if ( ! $post_id ) {
				MfLog::log('error', 'Error creating new post in WP ('. $mf_id .'): wp_insert_post returned 0' );
                return false;
            }

            update_post_meta( $post_id, 'coupon_amount', '' );
            update_post_meta( $post_id, 'individual_use', 'no' );
            update_post_meta( $post_id, 'product_ids', '' );
            update_post_meta( $post_id, 'exclude_product_ids', '' );
            update_post_meta( $post_id, 'usage_limit', '' );
            update_post_meta( $post_id, 'expiry_date', '' );
            update_post_meta( $post_id, 'apply_before_tax', 'yes' );
            update_post_meta( $post_id, 'free_shipping', 'no' );

            update_post_meta( $post_id, 'mf_id', $mf_id );
            update_post_meta( $post_id, 'mf_type', 'campaign' );
			MfLog::log('debug', 'New post created in WP ('. $mf_id .'; '. $post_type .'; '. $name .')' );
        }
        
        return $post_id;
    }

    /**
     * Gets post id that has unique mf id.
     * 
     * @param string $mf_id Unique mf id
     * @param string $post_type Post Type
     * @return int|bool Post id on success, else false
     */
    protected function get_post_id( $mf_id, $post_type ) {
        $posts = get_posts([
            'post_type' => $post_type,
            'status' => 'all',
            'posts_per_page' => 1,
            'ignore_sticky_posts' => true,
            'meta_query' => [
                [
                    'key' => 'mf_id',
                    'value' => $mf_id,
                ],
                [
                    'key' => 'mf_type',
                    'value' => 'campaign',
                ],
            ],
        ]);

        if ( ! empty( $posts ) ) {
            return $posts[0]->ID;
        }

        return false;
    }

    /**
     * Deletes a post that has unique mf id. If it's not found then it does nothing
     * 
     * @param string $mf_id Unique mf id
     * @param string $post_type Post type
     * @return bool True on deleted, false on not found
     */
    protected function delete_post( $mf_id, $post_type ) {
        $post_id = $this->get_post_id( $mf_id, $post_type );

        if ( $post_id ) {
            $title = get_the_title( $post_id );
            wp_delete_post( $post_id );
	        wp_cache_delete( WC_Cache_Helper::get_cache_prefix( 'coupons' ) . 'coupon_id_from_code_' . $title, 'coupons' );
			MfLog::log('debug', 'Post deleted in WP ('. $mf_id .'; '. $post_type .'; '. $title .')' );
            return true;
        }

        return false;
    }
}
