<?php

defined('ABSPATH') || exit;

add_action('admin_head', 'Moreflo_Admin_Status::admin_styles');
add_action('admin_notices', 'Moreflo_Admin_Status::admin_notices_product');
add_action('woocommerce_variation_options', 'Moreflo_Admin_Status::admin_notice_edit_variation', 10, 3);
add_filter('manage_product_posts_columns', 'Moreflo_Admin_Status::manage_product_posts_columns', 20);
add_filter('manage_product_posts_custom_column', 'Moreflo_Admin_Status::manage_product_posts_custom_column', 10, 2);
add_action('init', 'Moreflo_Admin_Status::check_sync_cron_status');

class Moreflo_Admin_Status
{
    public static function admin_styles()
    {
        echo '<style>
			.fixed .column-apps_product_sync_status {
				text-align: center;
				width: 4%;
			}
			.mf-sync-ok {
				color: #7ad03a;
			}
			.mf-sync-error {
				color: #a44;
			}
		</style>';
    }

    /**
     * Sets sync status of product using Apps response
     * 
     * @param int $product_id Post Id
     * @param string $response Json response from Apps as a string 
     */
    public static function set_sync_status_from_response($product_id, $response)
    {
        $success = false;
        $error_message = null;

        try {
            $response_json = json_decode($response);
            $success = $response_json->list[0]->Success ? 'OK' : 'ERROR';
            if ('ERROR' == $success) {
                $error_message = $response_json->list[0]->ExtraInformation[0];
            }
        } catch (Exception $e) {
            $success = false;
            $error_message = __('Unknown error', 'wpmf');
        }

        update_post_meta($product_id, '_mf_sync_status', $success);
        update_post_meta($product_id, '_mf_sync_error_message', $error_message);
        update_post_meta($product_id, '_mf_sync_last_sync', date_i18n('Y-m-d H:i:s'));
    }

    /**
     * Sets sync status of product to ok
     * 
     * @param int $product_id Post Id
     */
    public static function set_sync_status_ok($product_id)
    {
        update_post_meta($product_id, '_mf_sync_status', 'OK');
        update_post_meta($product_id, '_mf_sync_error_message', '');
        update_post_meta($product_id, '_mf_sync_last_sync', date_i18n('Y-m-d H:i:s'));
    }

    public static function admin_notices_product()
    {
        if (!function_exists('get_current_screen')) {
            return;
        }

        if ('product' != get_current_screen()->id) {
            return;
        }

        global $post;
        if (!$post->ID) {
            return;
        }

        $statuses = [];
        $sync_status = self::get_product_sync_statuses($post->ID, $statuses);

        if ('ERROR' == $sync_status) {
            self::print_admin_notice_errors($statuses);
        }
    }

    public static function admin_notice_edit_variation($loop, $variation_data, $variation)
    {
        $statuses = [];
        $sync_status = self::get_product_sync_statuses($variation->ID, $statuses);

        if ('ERROR' == $sync_status) {
            self::print_admin_notice_errors($statuses);
        }
    }

    public static function manage_product_posts_columns($columns)
    {
        return array_merge($columns, ['apps_product_sync_status' => 'Apps']);
    }

    public static function manage_product_posts_custom_column($column_key, $post_id)
    {
        if ('apps_product_sync_status' == $column_key) {
            $statuses = [];
            $sync_status = self::get_product_sync_statuses($post_id, $statuses);

            $title = [];
            if ('OK' == $sync_status) {
                foreach ($statuses as $s) {
                    $title[] = sprintf(__('[%s] OK (id %d)', 'wpmf'), $s['last'], $s['id']);
                    if (count($title) > 2) {
                        $title[] = '..';
                        break;
                    }
                }

                printf(
                    '<span title="%s" class="mf-sync-ok"><span class="dashicons dashicons-yes-alt"></span></span>',
                    esc_attr(implode(', ', $title))
                );
            } elseif ('ERROR' == $sync_status) {
                foreach ($statuses as $i => $s) {
                    if ('ERROR' == $s['status']) {
                        $title[$i] = sprintf('[%s] %s', $s['last'], $s['error_message']);
                        if (count($title) > 3) {
                            $title[$i] .= '..';
                            break;
                        }
                    }
                }

                $message = implode(', ', $title);
                if (count($title) > 1) {
                    $message = __('Multiple errors. Edit product to view them', 'wpmf');
                }

                printf(
                    '<span title="%s" class="mf-sync-error"><span class="dashicons dashicons-warning"></span></span>',
                    esc_attr($message)
                );
            } else {
                printf('<span>&nbsp;</span>');
            }
        }
    }

    private static function print_admin_notice_errors($statuses)
    {
        $title = [];
        foreach ($statuses as $i => $s) {
            if ('ERROR' == $s['status']) {
                $title[$i] = sprintf('[%s] %s (id %d)', $s['last'], $s['error_message'], $s['id']);
                if (count($title) > 3) {
                    $title[$i] .= '..';
                    break;
                }
            }
        }

        printf(
            '<div class="%1$s"><p><strong>%2$s</strong></p><p>%3$s</p></div>',
            'notice notice-error',
            _n('MoreFlo Apps sync error', 'MoreFlo Apps sync errors', count($title), 'wpmf'),
            nl2br(esc_html(implode("\n", $title)))
        );
    }

    /**
     * @param int $post_id
     * @param array &$statuses Filled with all the statuses
     * @return string Status summary: 'OK', 'ERROR' or empty
     */
    private static function get_product_sync_statuses($post_id, &$statuses = [])
    {
        $sync_status = '';
        try {
            $product = wc_get_product($post_id);
            if (!is_a($product, 'WC_Product')) {
                return '';
            }

            // Get all statuses
            if (in_array($product->get_type(), ['simple', 'variation'])) {
                $statuses[] = self::get_product_sync_status($post_id);
            } elseif ('variable' == $product->get_type()) {
                $variations = self::get_product_variations($product);
                for ($i = 0; $i < count($variations); $i++) {
                    $statuses[$i] = self::get_product_sync_status($variations[$i]['variation_id']);
                }
            }

            // Get summary
            if (empty($statuses)) {
            } elseif (count(array_filter($statuses, function ($s) {
                return 'OK' == $s['status'];
            })) == count($statuses)) {
                $sync_status = 'OK';
            } elseif (count(array_filter($statuses, function ($s) {
                return 'ERROR' == $s['status'];
            })) > 0) {
                $sync_status = 'ERROR';
            }
        } catch (Exception $e) {
            $statuses = [];
            $sync_status = '';
        }

        return $sync_status;
    }

    private static function get_product_sync_status($post_id)
    {
        return [
            'id' => $post_id,
            'status' => get_post_meta($post_id, '_mf_sync_status', true),
            'error_message' => get_post_meta($post_id, '_mf_sync_error_message', true),
            'last' => get_post_meta($post_id, '_mf_sync_last_sync', true),
        ];
    }

    private static function get_product_variations($product)
    {
        try {
            return $product->get_available_variations();
        } catch (Exception $e) {
            return [];
        }
    }

	/**
	 * Get current status of Sync Cron.
	 *
	 * @return string If cron is running, return time to next sync. Otherwise, return 'inactive' or 'disabled' depending on status.
	 */
	public static function get_sync_cron_status(): string {
		$mf_settings       = get_option( 'moreflo_api_url' );
		$both_cron_enabled = wp_next_scheduled( 'mf_sync_articles' ) && wp_next_scheduled( 'mf_sync_orders' );

		if ( $both_cron_enabled && $time = wp_next_scheduled( MF_SYNC_ARTICLES ) ) {
			return human_time_diff( time(), $time );
		} elseif ( isset( $mf_settings['mf_sync_status'] ) && 'enabled' == $mf_settings['mf_sync_status'] ) {
			return 'inactive';
		} else {
			return 'disabled';
		}
	}

	public static function check_sync_cron_status() {
		if ( 'inactive' !== self::get_sync_cron_status() ) {
			return;
		}

		if ( get_option( MF_SYNC_CRON_CHECK_EMAIL ) ) {
			return;
		}

		$mail_sent = wp_mail(
			get_option( 'admin_email' ),
			__( 'MoreFlo Sync is inactive', 'wpmf' ),
			sprintf(
				__( 'Something has caused MoreFlo Sync to become inactive at %1$s, please visit %2$s to reactive it.', 'wpmf' ),
				'<a href="' . get_site_url() . '">' . get_site_url() . '</a>',
				'<a href="' . get_admin_url( null, 'admin.php?page=wp_moreflo_admin' ) . '">' . __( 'MoreFlo Sync settings', 'wpmf' ) . '</a>'
			),
			array( 'Content-Type: text/html; charset=UTF-8' )
		);

		if ( $mail_sent ) {
			update_option( MF_SYNC_CRON_CHECK_EMAIL, true );
		}
	}
}
