<?php

defined( 'ABSPATH' ) || exit;

if ( class_exists( 'Moreflo_Lock' ) ) {
    return;
}

/**
 * Used to lock operation to only run once
 */
class Moreflo_Lock {
    /**
     * @param string $slug Lock name using only lowercase and underscores
     */
    public function __construct( $slug ) {
        $this->key_time = 'mf_lock_'. get_current_blog_id() .'_time_'. $slug;
        $this->key_value = 'mf_lock_'. get_current_blog_id() .'_value_'. $slug;
        $this->value = wp_generate_uuid4();
    }

    /**
     * Creates a lock. This function will sleep for 50ms to ensure lock is not overwritten by other process.
     * 
     * @return bool True on successful lock, false on already locked
     */
    public function lock() {
        if ( $this->_is_any_lock() ) {
            // Other process uses this lock
            return false;
        }

        $this->_set_lock();

        if ( ! $this->_is_my_lock() ) {
            // Other process has overwritten this lock
            return false;
        }

        // Our lock is OK
        return true;
    }

    /**
     * Clears all locks of set name. Set by this process or otherwise.
     */
    public function clear_lock() {
        global $wpdb;

        $wpdb->delete( $wpdb->options, [ 'option_name' => $this->key_time ], [ '%s' ] );
        $wpdb->delete( $wpdb->options, [ 'option_name' => $this->key_value ], [ '%s' ] );
    }

    /**
     * Gets current lock value
     * 
     * @return false|string
     */
    public function get_lock_value() {
        global $wpdb;

        $inserted_time = $wpdb->get_var( $wpdb->prepare( "SELECT option_value FROM {$wpdb->options} WHERE option_name = %s ORDER BY option_id ASC", $this->key_time ) );
        $inserted_value = $wpdb->get_var( $wpdb->prepare( "SELECT option_value FROM {$wpdb->options} WHERE option_name = %s ORDER BY option_id ASC", $this->key_value ) );

        if ( is_null( $inserted_time ) || $inserted_time + 300 < time() ) {
            $this->clear_lock();
            return false;
        }

        return $inserted_value;
    }

    /**
     * Internal function
     */
    public function _is_any_lock() {
        return false !== $this->get_lock_value();
    }

    /**
     * Internal function
     */
    public function _set_lock() {
        global $wpdb;

        $inserted_time = $wpdb->replace(
            $wpdb->options,
            [
                'option_name' => $this->key_time,
                'option_value' => time(),
                'autoload' => 'no',
            ],
            [
                '%s',
                '%s',
                '%s',
            ]
        );

        $inserted_val = $wpdb->replace(
            $wpdb->options,
            [
                'option_name' => $this->key_value,
                'option_value' => $this->value,
                'autoload' => 'no',
            ],
            [
                '%s',
                '%s',
                '%s',
            ]
        );

        if ( ! $inserted_time || ! $inserted_val ) {
            return false;
        }

        usleep( rand( 10, 50 ) * 1000 ); // Sleep 10ms - 50ms

        return true;
    }

    /**
     * Internal function
     */
    public function _is_my_lock() {
        return $this->value === $this->get_lock_value();
    }

    public $key_time;
    public $key_value;
    public $value;
}
