<?php

defined( 'ABSPATH' ) || exit;

class MF_Sync_Groups_Sync implements MF_Sync_Sync {
    /**
     * Syncs groups to WooCommerce as product categories
     */
    public function sync( $page, $pageSize, $updatedSince ) {
		$api = new MfApi();
		$status = array('add' => 0, 'new' => 0, 'update' => 0, 'page' => 0, 'total' => 1);
		$params = array(
			'page' => $page,
			'pageSize' =>  $pageSize
		);

		$groups = $api->get_groups($params);

		if ( ! empty( $groups ) && ! empty( $groups['Total'] ) ) {
			$status['page'] = $groups['Page'];
			$status['total'] = $groups['Total'];
		}

		if ( ! empty( $groups['list'] ) && is_array( $groups['list'] ) ) {
			$status = $this->process_pulled_groups( $groups['list'], $status );
		}

		return $status;
    }

	public function process_pulled_groups( $list, $status ) {
		$list = $this->order_by_hierarchy( $list );

		if ( empty( $list ) ) {
			return $status;
		}

		foreach ( $list as $group ) {
			if ( empty( $group['Group'] ) ) {
				continue;
			}
			
			$action = $this->process_pulled_group( $group['Group'] );

			if($action !== false) {
				$status[ $action ]++;
			}
		}

		update_option( 'mf_sync_group_last_synced', time() );

		return $status;
	}

	public function process_pulled_group($group) {
		$return = false;

		MfLog::log(DEBUG, 'Processing group: ' . $group['Id']);

		$term_id = null;
		$args = [
			'parent' => 0,
			'name' => $group['Name'],
		];

		$current_terms = get_terms([
			'taxonomy' => 'product_cat',
			'hide_empty' => false,
			'meta_query' => [
				[
					'key' => 'mf_group_id',
					'value' => $group['Id'],
				],
			],
		]);

		if ( ! empty( $current_terms ) ) {
			$term_id = $current_terms[0]->term_id;
		}

		if ( $group['ParentId'] ) {
			$parent_terms = get_terms([
				'taxonomy' => 'product_cat',
				'hide_empty' => false,
				'meta_query' => [
					[
						'key' => 'mf_group_id',
						'value' => $group['ParentId'],
					],
				],
			]);
	
			if ( ! empty( $parent_terms ) ) {
				$args['parent'] = $parent_terms[0]->term_id;
			}
		}

		if ( is_null( $term_id ) ) {
			$term_parent = isset( $args['parent'] )
				? $args['parent']
				: null;
			$term_id = term_exists( $group['Name'], 'product_cat', $term_parent );
			if ( ! is_null( $term_id ) ) {
				$term_id = $term_id['term_id'];
				$return = 'update';
				MfLog::log(DEBUG, 'Re-using WC group '. $term_id .' for '. $group['Id']);
			} else {
				$insert_result = wp_insert_term( $group['Name'], 'product_cat', $args );
				if ( is_wp_error( $insert_result ) ) {
					MfLog::log(DEBUG, 'Error creating WC group ('. $group['Id'] .') : '. $insert_result->get_error_message());
					return false;
				}

				$term_id = $insert_result['term_id'];
				$return = 'add';
			}

			MfLog::log(DEBUG, 'Group created in WC: '. $group['Id']);
		} else {
			wp_update_term( $term_id, 'product_cat', $args );
			$return = 'update';

			MfLog::log(DEBUG, 'Group updated in WC: '. $group['Id']);
		}

		update_term_meta( $term_id, 'order', $group['DisplayOrder'] );
		update_term_meta( $term_id, 'mf_group_id', $group['Id'] );
		update_term_meta( $term_id, 'mf_last_updated', time() );

		MfLog::log(DEBUG, 'Processed group no: ' . $group['Id']);

		return $return;
	}

	function clean_groups($page) {
		$status = array('new' => 0, 'update' => 0, 'page' => $page, 'total' => 1);

		if ($page > 0) {
			// Already cleaned
			MfLog::log(DEBUG, 'Skipped double clean groups');
			return $status;
		}

		$last_group_import = get_option( 'mf_sync_group_last_synced', 0 );
		if ( $last_group_import < time() - MINUTE_IN_SECONDS ) {
			// Already cleaned
			MfLog::log(DEBUG, 'Skipped clean groups due to skipped group import');
			return $status;
		}


		MfLog::log(DEBUG, 'Processing clean groups');
		
		$outdated_terms = get_terms([
			'taxonomy' => 'product_cat',
			'hide_empty' => false,
			'fields' => 'ids',
			'meta_query' => [
				'relation' => 'AND',
				[
					'key' => 'mf_group_id',
					'compare' => 'EXISTS',
				],
				[
					'key' => 'mf_last_updated',
					'value' => time() - 5 * MINUTE_IN_SECONDS,
					'compare' => '<',
				],
			],
		]);

		if (!empty($outdated_terms)) {
			foreach ($outdated_terms as $term_id) {
				$mf_group_id = get_term_meta( $term_id, 'mf_group_id', true );
				wp_delete_term( $term_id, 'product_cat' );
				MfLog::log(DEBUG, 'Deleted product term: '. $mf_group_id);
			}
		}

		MfLog::log(DEBUG, 'Processed clean groups');

		return $status;
	}

	/**
	 * Orders group list with parents first
	 * 
	 * @param array $list
	 * @return array Ordered list
	 */
	public function order_by_hierarchy( $list ) {
		$ordered_list = $this->do_order_by_hierarchy( $list );

		$ordered_list_ids = array_map( function ( $i ) {
			return $i['Group']['Id'];
		}, $ordered_list );

		foreach ( $list as $item ) {
			if ( in_array( $item['Group']['Id'], $ordered_list_ids ) ) {
				continue;
			}

			$ordered_list[] = $item;
		}

		return $ordered_list;
	}

	private function do_order_by_hierarchy( $list, $parent_id = null ) {
		$matching_parent = array_filter( $list, function ( $item ) use ( $parent_id ) {
			return $item['Group']['ParentId'] === $parent_id;
		} );
		$not_matching_parent = array_filter( $list, function ( $item ) use ( $parent_id ) {
			return $item['Group']['ParentId'] !== $parent_id;
		} );

		if ( empty( $not_matching_parent ) ) {
			return $matching_parent;
		}

		if ( empty( $matching_parent ) ) {
			return [];
		}

		$ordered_children = [];
		foreach ( $matching_parent as $matching_parent_item ) {
			$ordered_children = array_merge(
				$ordered_children,
				$this->do_order_by_hierarchy( $not_matching_parent, $matching_parent_item['Group']['Id'] )
			);
		}

		return array_merge(
			$matching_parent,
			$ordered_children
		);
	}
}
