<?php

defined( 'ABSPATH' ) || exit;

class MF_Sync_Receipts_Factory {
    /**
     * Create receipt from WooCommerce order or refund order
     * 
     * @param WC_Order|WC_Order_Refund $order WooCommerce Order object
     * @param MF_Sync_Model_Account_Code[] $account_codes All available account codes
     * @param string $company_corp_no Company corp number, visible on receipt
     * @param string $company_name Company name, visible on receipt
     * @param int $customer_number Customer number
     * @param string $shipping_sku Apps shipping SKU
     * @param string[] $payment_method_accounts List of payment method accounts, keyed by payment method ids
     * @param string $shipping_account Shipping account
     * @param int $vat_account_0 Tax account 0%
     * @param int $vat_account_6 Tax account 6%
     * @param int $vat_account_12 Tax account 12%
     * @param int $vat_account_25 Tax account 25%
     * @param bool $do_stock_management If the receipt should be in charge of stock.
     * @return MF_Sync_Model_Receipt|WP_Error|false Receipt on success, else WP_Error
     */
    public static function create_receipt( $order, $account_codes, $company_corp_no, $company_name, 
                                            $customer_number, $shipping_sku, $payment_method_accounts, 
                                            $shipping_account, $vat_account_0, $vat_account_6, 
                                            $vat_account_12, $vat_account_25, $do_stock_management ) {
        $order_items = $order->get_items();
        if ( empty( $order_items ) ) {
            return new WP_Error( 'receipt_no_order_items', 'Order contains no items: '. $order->get_id() );
        }

        $order_references = [];
        $order_mf_id = get_post_meta( $order->get_id(), 'mf_id', true );
        $order_reference_index = null;
        if ( $order_mf_id ) {
            $order_reference_index = 0;
            $order_reference = self::create_order_reference([
                'Index' => $order_reference_index,
                'OrderNumber' => mf_get_apps_order_id( $order->get_id() ),
                'Origin' => 'AppsOrder',
                'OrderChangeState' => is_a( $order, 'WC_Order' )
                    ? 'Paid' // WC_Order
                    : 'Cancelled', // WC_Order_Refund
            ]);
            $order_references[] = $order_reference;
        }

        $parent_order = 'shop_order_refund' == $order->get_type()
            ? wc_get_order( $order->get_parent_id() )
            : null;

        // Make sure we can get apps order id
        if ( is_null( $parent_order ) ) {
            if ( is_null( mf_get_apps_order_id( $order->get_id() ) ) ) {
                return new WP_Error( 'receipt_missing_order_apps_id', 'Order has not correctly been synced to Apps: '. $order->get_id() );
            }
        } else {
            if ( is_null( mf_get_apps_order_id( $parent_order->get_id() ) ) ) {
                return new WP_Error( 'receipt_missing_order_apps_id_refund', 'Order for refund has not correctly been synced to Apps: '. $parent_order->get_id() );
            }
        }

        $order_payment_method = is_null( $parent_order )
            ? $order->get_payment_method()
            : $parent_order->get_payment_method();

        $payment_method_account = isset( $payment_method_accounts[ $order_payment_method ] )
            ? $payment_method_accounts[ $order_payment_method ]
            : null;

        if ( is_null( $payment_method_account ) ) {
            return new WP_Error( 'receipt_no_payment_method_account', 'No bookkeeping account set for payment method: '. $order->get_payment_method() );
        }

        $woocommerce_calc_taxes_enable = get_option('woocommerce_calc_taxes');
        if ( 'yes' != $woocommerce_calc_taxes_enable ) {
            return new WP_Error( 'receipt_taxes_off', 'Taxes must be enabled' );
        }

        $coupon_name = __( 'Discount', 'wpmf' );

        $articles = [];
        foreach ( $order_items as $order_item ) {
            if ( ! is_a( $order_item, 'WC_Order_Item_Product' ) ) {
                return new WP_Error( 'receipt_unknown_order_item', 'Unknown order item: '. $order_item->get_name() );
            }

            $articles[] = self::create_article( $order_item, $order, $order_reference_index, $account_codes, $coupon_name, $parent_order );
        }

        $shipping = self::create_shipping( $order, $shipping_sku, $shipping_account );
        if ( ! is_null( $shipping ) ) {
            $articles[] = $shipping;
        }

        if ( in_array( null, $articles, true ) ) {
            return new WP_Error( 'receipt_articles_missing', __( 'Articles from order are missing', 'wpmf' ) );
        }

        $vat = [];
        $tax_items = $order->get_taxes();
        if ( ! empty( $tax_items ) ) {
            foreach ( $tax_items as $tax_item ) {
                if ( ! $tax_item->get_tax_total() && ! $tax_item->get_shipping_tax_total() ) {
                    // Ignore empty tax totals
                    continue;
                }

                $vat_row = self::create_vat( $tax_item, $articles, $vat_account_0, $vat_account_6, 
                    $vat_account_12, $vat_account_25 );
                if ( is_null( $vat_row ) ) {
                    continue;
                }

                $vat[] = $vat_row;
            }
        }

        $internal_note = null;
        if ( 'shop_order_refund' == $order->get_type() ) {
            $internal_note = $order->get_reason();
        }

        $receipt_date = is_null( $parent_order )
            ? $order->get_date_paid()
            : $order->get_date_created();

        if ( is_null( $receipt_date ) ) {
            return new WP_Error( 'order_not_paid', __( 'Order has not been paid', 'wpmf' ) );
        }

        $merchant = self::create_merchant([
            'CorporateIdentityNumber' => $company_corp_no,
            'Name' => $company_name,
            'AddressLine1' => WC()->countries->get_base_address(),
            'AddressLine2' => WC()->countries->get_base_address_2(),
            'ZipCode' => WC()->countries->get_base_postcode(),
            'City' => WC()->countries->get_base_city(),
            'Country' => WC()->countries->get_base_country(),
            'Phone' => null,
            'HomePage' => get_site_url(),
            'Email' => null,
        ]);

        $customers = [
            self::create_customer([
                'CustomerNumber' => $customer_number,
                'CompanyName' => is_null( $parent_order )
                    ? $order->get_billing_company()
                    : $parent_order->get_billing_company(),
                'Fullname' => is_null( $parent_order )
                    ? $order->get_formatted_billing_full_name()
                    : $parent_order->get_formatted_billing_full_name(),
                'CustomerType' => 'Customer',
            ]),
        ];

        $payments = [
            self::create_payment([
                'Name' => __( 'Online Payment', 'wpmf' ),
                'Amount' => $order->get_total(),
                'PaymentType' => 'NamedPayment',
                'TransactionId' => is_null( $parent_order )
                    ? $order->get_transaction_id()
                    : $parent_order->get_transaction_id(),
                'TransactionOrigin' => $order_payment_method,
            ]),
        ];

        $voucher = self::create_voucher( $order, $articles, $vat, $payment_method_account );
        
        return new MF_Sync_Model_Receipt([
            'ReceiptNumber' => $order->get_id(),
            'Origin' => 'WooCommerce - '. parse_url( get_site_url(), PHP_URL_HOST ),
            'ReceiptType' => 'Online',
            'InternalNote' => $internal_note,
            'TransactionDate' => $receipt_date->format('Y-m-d\TH:i:s\.\0\0\0\Z'),
            'TotalPriceIncVat' => floatval( $order->get_total() ),
            'TotalPriceExVat' => round( floatval( $order->get_total() ) - floatval( $order->get_total_tax() ), 2 ),
            'TotalVat' => floatval( $order->get_total_tax() ),
            'DoStockMgmt' => $do_stock_management,
            'Merchant' => $merchant,
            'Customers' => $customers,
            'Articles' => $articles,
            'Payments' => $payments,
            'Orders' => $order_references,
            'Discounts' => [],
            'Vat' => $vat,
            'Voucher' => $voucher,
        ]);
    }

    /**
     * Create article from order item product.
     * 
     * @param WC_Order_Item_Product $order_item
     * @param WC_Order|WC_Order_Refund $order Order that the item refers to.
     * @param int $order_reference_index Reference to order
     * @param MF_Sync_Model_Account_Code[] $account_codes
     * @param string $coupon_name Coupon name is only used if article is discounted.
     * @param WC_Order|null $parent_order Parent order if refund
     * @return MF_Sync_Model_Article|null Article on success, else null.
     */
    public static function create_article( $order_item, $order, $order_reference_index, $account_codes, $coupon_name, $parent_order = null ) {
        $product = $order_item->get_product();

        if ( ! is_a( $product, 'WC_Product' ) ) {
            return null; // Prouct cannot be found, abort
        }

        $vat = is_null( $parent_order )
            ? MF_Sync_Receipts_Helper::get_order_row_tax_rate( $order_item, $order )
            : MF_Sync_Receipts_Helper::get_order_row_tax_rate( $order_item, $parent_order );

        $price = (float) $product->get_regular_price();
        if ( 'no' == get_option('woocommerce_prices_include_tax') ) {
            $tax_amount = $price * ( $vat / 100 );
            $price = $price + $tax_amount;
        }

        $salesprice = $order->get_item_total( $order_item, true, true );
        $dis = 0;
        if ( $salesprice < $price ) {
            $dis = abs( $price - $order->get_item_total( $order_item, true, false ) );
        }

        $attributes = [];
        $serial_number = MF_Sync_Receipts_Helper::get_serial_number( $product );
        if ( $serial_number ) {
            $attributes[] = [
                'AttrbuteType' => 'SerialNumber',
                'Value' => $serial_number,
            ];
        }

        $quantity = $order_item->get_quantity();
        if ( $quantity < 0 ) {
            // Refund rows should contain positive quantities and negative unit prices
            $quantity *= -1;
            $price *= -1;
        }

        $discounts = [];
        if ( $dis ) {
            $discounts[] = [
                'DiscountType' => 'amount',
                'Amount' => mf_wc_round( $dis * $quantity, 2 ),
                'Description' => $coupon_name,
                'Percentage' => null, // null during type 'amount'
            ];
        }

        $article_name = $order_item->get_name();
        if ( is_a( $product, 'WC_Product_Variation' ) ) {
            // WooCommerce sometimes doesn't add all the proper attributes to the name of the variation. This will force add them to the title
            $possible_attributes = array_keys( $product->get_attributes() );
            $title_base = get_post_field( 'post_title', $product->get_parent_id() );
            $meta_data = $order_item->get_formatted_meta_data();
            $meta_data = array_filter( $meta_data, function ( $v ) use ( $possible_attributes ) {
                return in_array( $v->key, $possible_attributes );
            } );

            if ( ! empty( $meta_data ) ) {
                $meta_data = array_map( function ( $v ) {
                    return $v->value;
                }, $meta_data );

                $title_separator = ' - ';
                $attribute_separator = ', ';
                $title_suffix = trim( strip_tags( implode( $attribute_separator, $meta_data ) ) );
                if ( $title_suffix && $title_base ) {
                    if ( $title_base == $article_name ) {
                        // Attributes are missing from article name entirely
                        $article_name = $article_name . $title_separator . $title_suffix;
                    } else {
                        // Some attributes are in the title
                        $article_name = $article_name  . $attribute_separator . $title_suffix;
                    }
                }
            }
        }

        return new MF_Sync_Model_Article([
            'OrderIndex' => $order_reference_index,
            'ArticleNumber' => $product->get_sku(),
            'Name' => $article_name,
            'Quantity' => $quantity,
            'UnitPrice' => mf_wc_round(
                $order->get_item_total( $order_item, true, false ),
                2
            ),
            'UnitPriceWithoutDiscounts' => mf_wc_round(
                $price,
                2
            ),
            'VatRate' => $vat,
            'TotalPriceIncVat' => floatval( $order->get_line_total( $order_item, true, true ) ),
            'TotalPriceExVat' => floatval( $order->get_line_total( $order_item, false, true ) ),
            'TotalVat' => floatval( $order_item->get_total_tax() ),
            'BookkeepingAccount' => empty( $account_codes )
                ? ''
                : MF_Sync_Receipts_Helper::bookkeeping_account_product( $account_codes, $product ),
            'Discounts' => $discounts,
            'Attributes' => $attributes,
            'type' => 'product',
            'taxes_totals' => $order_item->get_taxes(),
        ]);
    }

    /**
     * @param WC_Order_Item_Tax $tax_item
     * @param MF_Sync_Model_Article[] $articles
     * @param int $vat_account_0 Tax account 0%
     * @param int $vat_account_6 Tax account 6%
     * @param int $vat_account_12 Tax account 12%
     * @param int $vat_account_25 Tax account 25%
     * @return MF_Sync_Model_Vat|null
     */
    public static function create_vat( $tax_item, $articles, $vat_account_0, $vat_account_6, 
        $vat_account_12, $vat_account_25 ) {
        $total_amount = 0.0;
        $total_vat = 0.0;
        foreach ( $articles as $article ) {
            $article_tax_total = $article->get_tax_total( $tax_item->get_rate_id() );
            if ( is_null( $article_tax_total ) ) {
                continue;
            }

            $total_amount += $article->get_total_price_inc_vat();
            $total_vat = round( $total_vat + floatval( $article_tax_total ), 2 ); // Fix binary/decimal conversion
        }
        if ( ! $total_amount ) {
            return null;
        }

        $vat_rate = $tax_item->get_rate_percent('edit');
        $vat_account = null;
        if ( 0 == $vat_rate ) {
            $vat_account = $vat_account_0;
        } elseif ( 6 == $vat_rate ) {
            $vat_account = $vat_account_6;
        } elseif ( 12 == $vat_rate ) {
            $vat_account = $vat_account_12;
        } elseif ( 25 == $vat_rate ) {
            $vat_account = $vat_account_25;
        }

        if ( is_null( $vat_account ) ) {
            return null;
        }

        return new MF_Sync_Model_Vat([
            'VatRate' => $vat_rate,
            'TotalVat' => $total_vat,
            'TotalAmount' => $total_amount,
            'BookkeepingAccount' => $vat_account,
        ]);
    }

    /**
     * @param WC_Order $order
     * @param string $shipping_sku
     * @param string $shipping_account
     * @return MF_Sync_Model_Article|null Article on success, null if no shipping.
     */
    public static function create_shipping( $order, $shipping_sku, $shipping_account ) {
        $shipping_tax_rate = 0;
        $shipping_tax = floatval( $order->get_shipping_tax() );
        $shipping_total = floatval( $order->get_shipping_total() ) + $shipping_tax;
        $taxes_totals = null;

        if ( $shipping_tax ) {
            $tax_lines = $order->get_items( 'shipping' );
            if ( ! empty( $tax_lines ) ) {
                foreach ( $tax_lines as $tax_line ) {
                    $taxes_totals = $tax_line->get_taxes();
                    $shipping_tax_total = round(
                        array_reduce(
                            $tax_line->get_taxes()['total'],
                            function ( $carry, $item ) {
                                return $carry + floatval( $item );
                            }
                        ),
                        2
                    );
                    $shipping_tax_rate = floatval( ceil( 100 * ( ( ( floatval( $order->get_shipping_total() ) + floatval( $shipping_tax_total ) ) / floatval( $order->get_shipping_total() ) ) - 1 ) ) );
                }
            }
        }

        if ( ! $shipping_total > 0 ) {
            return null;
        }

        return new MF_Sync_Model_Article([
            'ArticleNumber' => $shipping_sku,
            'Name' => __('Shipping', 'wpmf'),
            'Quantity' => 1,
            'UnitPrice' => $shipping_total,
            'VatRate' => $shipping_tax_rate,
            'TotalPriceIncVat' => floatval( $shipping_total ),
            'TotalPriceExVat' => floatval( $order->get_shipping_total() ),
            'TotalVat' => $shipping_tax,
            'BookkeepingAccount' => $shipping_account,
            'Discounts' => [],
            'Attributes' => [],
            'type' => 'shipping',
            'taxes_totals' => $taxes_totals,
        ]);
    }

    /**
     * @param WC_Order_Item_Coupon $coupon
     */
    public static function create_discount( $coupon ) {
        return new MF_Sync_Model_Receipt_Discount([
            'DiscountType' => 'Amount',
            'Amount' => floatval( $coupon->get_discount('edit') ),
            'Description' => $coupon->get_name(),
            'Percentage' => null,
        ]);
    }

    /**
     * @param WC_Order|WC_Order_Refund $order
     * @param MF_Sync_Model_Article[] $articles
     * @param MF_Sync_Model_Vat[] $vats
     * @param string $payment_method_account
     * @return MF_Sync_Model_Voucher
     */
    public static function create_voucher( $order, $articles, $vats, $payment_method_account ) {
        $parent_order = 'shop_order_refund' == $order->get_type()
            ? wc_get_order( $order->get_parent_id() )
            : null;

        $rows = [
            [
                'AccountNumber' => $payment_method_account,
                'DebitAmount' => floatval( $order->get_total() ),
                'CreditAmount' => 0,
            ],
        ];

        foreach ( $articles as $article ) {
            if ( 'product' == $article->get_type() && $article->get_total_price_ex_vat() != 0 ) {
                $rows[] = [
                    'AccountNumber' => $article->get_bookkeeping_account(),
                    'DebitAmount' => 0,
                    'CreditAmount' => $article->get_total_price_ex_vat(),
                ];
            } elseif ( 'shipping' == $article->get_type() && $article->get_total_price_ex_vat() != 0 ) {
                $rows[] = [
                    'AccountNumber' => $article->get_bookkeeping_account(),
                    'DebitAmount' => 0,
                    'CreditAmount' => $article->get_total_price_ex_vat(),
                ];
            }
        }

        if ( ! empty( $vats ) ) {
            foreach ( $vats as $vat ) {
                if ( $vat->get_total_vat() != 0 ) {
                    $rows[] = [
                        'AccountNumber' => $vat->get_bookkeeping_account(),
                        'DebitAmount' => 0,
                        'CreditAmount' => $vat->get_total_vat(),
                    ];
                }
            }
        }

        // Swap Debit and Credit for negative values
        $rows = array_map( function ( $row ) {
            if ( $row['DebitAmount'] < 0 ) {
                $row['CreditAmount'] += abs( $row['DebitAmount'] );
                $row['DebitAmount'] = 0;
            }
            if ( $row['CreditAmount'] < 0 ) {
                $row['DebitAmount'] += abs( $row['CreditAmount'] );
                $row['CreditAmount'] = 0;
            }
            
            return $row;
        }, $rows );

        // Transform into voucher rows
        $rows = array_map( function ( $row ) {
            return new MF_Sync_Model_Voucher_Row( $row );
        }, $rows );

        $voucher_date = is_null( $parent_order )
            ? $order->get_date_paid()
            : $order->get_date_created();

        $voucher_text = is_null( $parent_order )
            ? sprintf( __( 'WooCommerce Order #%s', 'wpmf' ), mf_get_apps_order_id( $order->get_id() ) )
            : sprintf( __( 'WooCommerce refund (%d) of Order #%s', 'wpmf' ), $order->get_id(), mf_get_apps_order_id( $parent_order->get_id() ) );
        
        return new MF_Sync_Model_Voucher([
            'State' => 'Pending',
            'ReferenceNumber' => is_null( $parent_order )
                ? $order->get_order_number()
                : $parent_order->get_order_number() .'-refund-'. $order->get_id(),
            'ReferenceType' => 'WebOrder',
            'Rows' => $rows,
            'VoucherDate' => is_null( $voucher_date )
                ? null
                : $voucher_date->format( 'Y-m-d\\TH:i:s' ),
            'VoucherText' => $voucher_text,
        ]);
    }

    /**
     * @param array $data
     * @return MF_Sync_Model_Voucher_Row
     */
    public static function create_voucher_row( $data ) {
        return new MF_Sync_Model_Voucher_Row( $data );
    }

    /**
     * @param array $data
     * @return MF_Sync_Model_Merchant
     */
    public static function create_merchant( $data ) {
        return new MF_Sync_Model_Merchant( $data );
    }

    /**
     * @param array $data
     * @return MF_Sync_Model_Customer
     */
    public static function create_customer( $data ) {
        return new MF_Sync_Model_Customer( $data );
    }

    /**
     * @param array $data
     * @return MF_Sync_Model_Payment
     */
    public static function create_payment( $data ) {
        return new MF_Sync_Model_Payment( $data );
    }

    /**
     * @param array $data
     * @return MF_Sync_Model_Account_Code
     */
    public static function create_account_code( $data ) {
        return new MF_Sync_Model_Account_Code( $data );
    }

    /**
     * @param array $data
     * @return MF_Sync_Model_OrderReference
     */
    public static function create_order_reference( $data ) {
        return new MF_Sync_Model_OrderReference( $data );
    }
}
